<?php

namespace HighwayPro\App\Data\Model\UrlViews;

use Carbon\Carbon;
use HighWayPro\App\Data\Schema\DestinationTable;
use HighWayPro\App\Data\Schema\UrlViewTable;
use HighWayPro\App\HighwayPro\Time\PeriodOfTime;
use HighWayPro\Original\Collections\Collection;
use HighWayPro\Original\Data\Model\Gateway;
use HighwayPro\App\Data\Model\Destinations\Destination;
use HighwayPro\App\Data\Model\UrlViews\UrlView;
use HighwayPro\App\Data\Model\Urls\Url;
use HighwayPro\App\HighwayPro\System\Date;
use HighwayPro\Original\Characters\IntegerManager;
use HighwayPro\Original\Characters\StringManager;

Class UrlViewGateway extends Gateway
{
    const NA_KEY = '__na__';
    const DATE_RANGE_FORMAT = 'Y-m-d';

    protected function model()
    {
        return [
            'table' => new UrlViewTable,
            'domain' => UrlView::class
        ];
    }

    public function getFromDestinationId($destinationId)
    {
        return $this->createCollection(
                    (array) $this->driver->get("SELECT * FROM {$this->table->getName()} WHERE destination_id = ?", [$destinationId])
                );
    } 

    public function saveView(UrlView $urlView)
    {
        return $this->driver->execute(
            "INSERT INTO {$this->table->getName()} ({$this->getFields($urlView)}) 
             VALUES({$this->getValuesAsMark($urlView)})", 
            $urlView->getAvailableValues()
            );
    }

    public function getTotalNumberOfRows($urlId = null, PeriodOfTime $periodOfTime = null)
    {
        (string) $optionallyFilteredById = '';
        (string) $optionallyFilteredByDateRange = '';

        if (!is_null($urlId)) {
            $optionallyFilteredById = "AND {$this->getFilteredByIdCondition()}";
        }

        if ($periodOfTime instanceof PeriodOfTime) {
            $optionallyFilteredByDateRange = "AND (date BETWEEN ? and ?)";
        }

        (array) $result = (array) (new Collection(
            (array) $this->driver->get(
                "
                    SELECT COUNT(*) as total 
                    FROM {$this->table->getName()}
                    WHERE 1 = 1 {$optionallyFilteredById}
                                {$optionallyFilteredByDateRange}
                ",
                array_filter([
                    $urlId,
                    $optionallyFilteredByDateRange? $periodOfTime->getStartDateFormattedWithTimestamp() : null,
                    $optionallyFilteredByDateRange? $periodOfTime->getEndDateFormattedWithTimestamp() : null
                ])
            )
        ))->first();

        return $this->getTotal($result);
    }

    /**
     *  $dateData must be an array containing a date range
     *  with 'from' and 'to' items as YYYY-MM-DD date strings
     * 
     * @param  Array  $dateData 
     * @return integer           
     */
    public function getTotalNumberOfRowsInPeriodWithDifference(PeriodOfTime $periodOfTime, 
                                                               PeriodOfTime $periodToCompare, 
                                                               $urlId = null)
    {
        (integer) $total = $this->getTotalNumberOfRowsInPeriod($periodOfTime, $urlId);
        (integer) $totalFromPeriodToCompare = $this->getTotalNumberOfRowsInPeriod($periodToCompare, $urlId);

        return new Collection([
            'total' => $total,
            'difference' => [
                'total' => $totalFromPeriodToCompare,
                'difference' => $total - $totalFromPeriodToCompare,
                'percentage' => (($total - $totalFromPeriodToCompare) / ($totalFromPeriodToCompare?: 1)) * 100
            ]
        ]);
    }

    public function getTotalNumberOfRowsInPeriod(PeriodOfTime $periodOfTime, $urlId = null)
    {
        (string) $optionallyFilteredById = '';

        if (!is_null($urlId)) {
            $optionallyFilteredById = "AND ({$this->getFilteredByIdCondition()})";
        }

        (array) $result = (array) (new Collection(
            (array) $this->driver->get(
                "
                    SELECT count(*) as total FROM {$this->table->getName()}
                    WHERE (date BETWEEN ? AND ?) {$optionallyFilteredById}
                ",
                array_filter([
                    $periodOfTime->getStartDateFormattedWithTimestamp(),
                    $periodOfTime->getEndDateFormattedWithTimestamp(),
                    $urlId
                ])
            )
        ))->first();

        return $this->getTotal($result);
    }

    /**
     * TimeStamp needed for testing
     */
    public function getFromPastDays($numberOfDays, $timeStamp = null)
    {
        (object) $date = $this->getDateBeforeANumberOfDays($numberOfDays, $timeStamp);

        return $this->createCollection(
            (array) $this->driver->get(
                "
                    SELECT * FROM {$this->table->getName()}
                    WHERE date >= ?
                ",
                [
                    $date->get('daysAgoDate'),
                ]
            )
        );
    }   

    public function getCountInPeriod(PeriodOfTime $periodOfTime, $urlId = null)
    {
        (string) $optionallyFilteredById = '';

        if ($urlId) {
            $optionallyFilteredById = "AND ({$this->getFilteredByIdCondition()})";
        }

        // Here the query is only returning results ONLY from the days there were rows
        // for example, if we are selecting jan 1 to jan 31, the query MAY NOT necessarily
        // return 31 rows, if there were rows in the 2nd and then no more rows were inserted until the 7th, then we'll have 
        // 26 rows instead of 31. We need to pad the remaining rows.
        (string) $OPTIONALLYBYHOUR = $periodOfTime->unitIs('hours')? ',HOUR(date)' : '';
        (string) $format = $periodOfTime->unitIs('hours')? Date::FORMAT : Date::FORMAT_NO_TIMESTAMP;

        (object) $scatteredPossiblyIncompleteDays = (new Collection(
            (array) $this->driver->get(
                "
                    SELECT COUNT(*) as total, 
                               DATE_FORMAT(date,'%%d') as day, 
                               DATE_FORMAT(date,'%%m') as month, 
                               YEAR(date) as year, 
                               date
                    FROM highwaypro_url_views
                    WHERE (date BETWEEN ? AND ?) {$optionallyFilteredById}
                    GROUP BY YEAR(date), MONTH(date), DAY(date) $OPTIONALLYBYHOUR
                    ORDER BY total DESC
                ",
                array_filter([
                    $periodOfTime->getStartDate()->format(Date::FORMAT),
                    $periodOfTime->getEndDate()->format(Date::FORMAT),
                    $urlId
                ])
            )
        ))->mapWithKeys(function($row) use ($format) {
            $row = (object) $row;
            $row->total = (integer) $row->total;
            $row->year = (integer) $row->year;
            return [
                'key' => date($format, strtotime($row->date)),
                'value' => $row
            ];
        });

        (object) $daysOrHours = $periodOfTime->unitIs('hours')? $periodOfTime->getHoursFromDateRange() : $periodOfTime->getDaysFromDateRange();

        (integer) $index = -1;
        $daysOrHoursWithData = $daysOrHours->map(function(Date $date) use (&$index, $scatteredPossiblyIncompleteDays, $periodOfTime, $format) {      
            // here is where we'll group results, if results are days but format is weeks, then we'll group every 7 dyas
            // if result is in months, then every month, etc
            // perhaps in a different callback funtion before calling ->filter() (down below)
            $index++;
            $data = (
                $periodOfTime->unitIs('hours')? $scatteredPossiblyIncompleteDays->find(function($row) use ($date) {
                    return date(Date::FORMAT_HOUR_ONLY, strtotime($row->date)) === $date->format(Date::FORMAT_HOUR_ONLY);
                }) : $scatteredPossiblyIncompleteDays->get($date->format($format))
            )?: (object) ['total' => 0];

            return (object) array_merge(
                [
                    'total' => $data->total,
                    "day"=> $date->format(Date::FORMAT_DAY_ONLY),
                    "date"=> $date->startOfHour()->format(Date::FORMAT),
                ]
            );
        })->filter();

        /*
        if (in_array($periodOfTime->getUnit(), ['hours', 'days'])) {
            return $daysOrHoursWithData;
        }*/

        // [Date $from<start of week>, $to<end of week>]
        // [Date $from<start of week>, $to<end of week>]
        // [Date $from<start of week>, $to<end of week>]
        // note that the lenght of the first level aray is the expected lenght (PeriodOfTime::getLength())
        (integer) $index = -1;
        return $periodOfTime->getDateRanges()->map(function($range) use (&$index, $daysOrHoursWithData, $OPTIONALLYBYHOUR, $periodOfTime) {
            (object) $from = $range instanceof Date? $range : $range->get('from');
            (object) $to =   $range instanceof Date? $range : $range->get('to');

            (integer) $count = (integer) $daysOrHoursWithData->filterAndRemove(function($data) use ($from, $to) {
                return Date::createFromString($data->date)->between($from, $to);
            })->reduce(function($count, $data){
                return $count + (integer) $data->total;
            }, $initial = 0);

            $index++;
            (string) $labelFormat = $periodOfTime->unitIs('hours')? 'g A' : 'D, M jS';
            (string) $dateFormat = in_array($periodOfTime->getUnit(), ['hours', 'days'])? $from->format($labelFormat) : "{$from->format($labelFormat)} -> {$to->format($labelFormat)}";

            return (object) [
                'count' => $count,
                'dateRange' => (object) [
                    'from' => (object) [
                        'date' => $from->format(Date::FORMAT),
                        'dateNoTimeStamp' => $from->format(Date::FORMAT_NO_TIMESTAMP),
                        'year' => $from->format(Date::FORMAT_YEAR_ONLY),
                        'month' => $from->format(Date::FORMAT_MONTH_ONLY),
                        'day' => $from->format(Date::FORMAT_DAY_ONLY),
                        'hour' => $from->format(Date::FORMAT_HOUR_ONLY),
                    ],
                    'to' => (object) [
                        'date' => $to->format(Date::FORMAT),
                        'dateNoTimeStamp' => $to->format(Date::FORMAT_NO_TIMESTAMP),
                        'year' => $to->format(Date::FORMAT_YEAR_ONLY),
                        'month' => $to->format(Date::FORMAT_MONTH_ONLY),
                        'day' => $to->format(Date::FORMAT_DAY_ONLY),
                        'hour' => $to->format(Date::FORMAT_HOUR_ONLY),
                    ],
                ],
                'labels' => (object) [
                    'date' => $dateFormat
                ],
                'offset' => $index
            ];

        });
    }

    public function getCountFromPastDays($numberOfDays, $urlId,  $timeStamp = null)
    {
        (string) $optionallyFilteredById = '';

        if (!is_null($urlId)) {
            $optionallyFilteredById = "AND ({$this->getFilteredByIdCondition()})";
        }

        (object) $date = $this->getDateBeforeANumberOfDays($numberOfDays, $timeStamp);

        (object) $rows = new Collection(
            (array) $this->driver->get(
                "
                    SELECT COUNT(*) as total, 
                           DAY(date) as day, 
                           MONTH(date) as month, 
                           YEAR(date) as year, 
                           date
                    FROM highwaypro_url_views
                    WHERE (date >= ?) {$optionallyFilteredById}
                    GROUP BY DAY(date)
                    ORDER BY total DESC
                ",
                array_filter([
                    $date->get('daysAgoDate'),
                    $urlId
                ])
            )
        );   

        return $date->get('daysRange')->map(function($day) use ($rows, $date) {
            (object) $rowForThatDay = $rows->filter(function($view) use ($day) {
                $view = (object) $view;

                return ((integer) $view->day) === ((integer) $day->get('day')) 
                        && 
                       ((integer) $view->month) === ((integer) $day->get('month'));
            });

            (string) $_day = $rowForThatDay->haveAny()? 
                            $rowForThatDay->first()['day'] : 
                            $day->get('day');

            $_day = IntegerManager::create($_day)->twoDigit();

            (string) $month = $rowForThatDay->haveAny()? 
                                $rowForThatDay->first()['month'] : 
                                $day->get('month');

            $month = IntegerManager::create($month)->twoDigit();

            (integer) $year = $rowForThatDay->haveAny()? 
                                $rowForThatDay->first()['year'] : 
                                $day->get('year');

            return (object) [
                "total"=> (integer) ($rowForThatDay->haveAny()? $rowForThatDay->first()['total'] : 0),
                "day"=> $_day,
                "month"=> $month,
                "year"=> (integer) $year,
                "date"=> "{$year}-{$month}-{$_day}",
                "offset" => $day->get('offset')
            ];
 
        });
    }

    public function getTotalCountByField($field, PeriodOfTime $periodOfTime = null, $urlId = null)
    {
        (string) $optionallyFilteredById = '';
        (string) $optionallyFilteredByDateRange = '';

        if (!is_null($urlId)) {
            $optionallyFilteredById = "AND ({$this->getFilteredByIdCondition()})";
        }

        if (!is_null($periodOfTime)) {
            $optionallyFilteredByDateRange = "AND (date BETWEEN ? AND ?)";
        }

        $field = (string) (new StringManager($field))->getAlphanumeric();

        (object) $rows = new Collection(
            (array) $this->driver->get(
                "
                    SELECT COUNT(*) as total, 
                           {$field}
                    FROM highwaypro_url_views
                    WHERE 1=1 {$optionallyFilteredById}
                              {$optionallyFilteredByDateRange}
                    GROUP BY {$field}
                    ORDER BY total DESC
                ",
                array_filter([
                    $urlId,
                    $periodOfTime? $periodOfTime->getStartDateFormattedWithTimestamp() : null,
                    $periodOfTime? $periodOfTime->getEndDateFormattedWithTimestamp() : null
                ])
            )
        );   

        (integer) $totalRows = $rows->reduce(function($total, $currentRow){
            return $total + $currentRow['total'];
        });

        (object) $rowsWithPercentage = $rows->map(function($row) use ($totalRows, $field) {
            $row = (object) $row;
            if ((string) $row->{$field} == '') $row->{$field} = static::NA_KEY;

            $row->total = (integer) $row->total;
            $row->percentage = (($row->total * 100) / $totalRows);

            return $row;
        });

        (object) $emptyValuesRow = $rowsWithPercentage->filter(function($row) use ($field) {
            return $row->{$field} === static::NA_KEY;
        })->reduce(function($previouslyProcessedRowsValue, $currentRow) {
            if (!is_object($previouslyProcessedRowsValue)) {
                return $currentRow;
            }

            $previouslyProcessedRowsValue->total += $currentRow->total;
            $previouslyProcessedRowsValue->percentage += $currentRow->percentage;

            return $previouslyProcessedRowsValue;
        });

        if (is_object($emptyValuesRow)) {
            return $rowsWithPercentage->filter(function($row) use ($field) {
                return $row->{$field} !== static::NA_KEY;
            })->push($emptyValuesRow)
              ->resetKeys();
        }

        return $rowsWithPercentage;
    }
    
    public function GetWithId($id)
    {
        return $this->createCollection(
                    (array) $this->driver->get("SELECT * FROM {$this->table->getName()} WHERE id = ?", [$id])
                )->first();
    }   

    public function getDateBeforeANumberOfDays($numberOfDays, $timeStamp = null)
    {
        (integer) $timeStamp = is_null($timeStamp)? strtotime("now") : $timeStamp; 
        (integer) $pastDaysTimeStamp = strtotime("-{$numberOfDays} days",  $timeStamp);

        (string) $daysAgoDate = date(
            UrlView::DATE_FORMAT, 
            $pastDaysTimeStamp
        );   

        (string) $currentDay = date("d", $timeStamp);
        (string) $currentMonth = date("m", $timeStamp);

        (string) $pastDay = date("d", $pastDaysTimeStamp);
        (string) $pastDayMonth = date("m", $pastDaysTimeStamp);
        (string) $pastDayYear = date("Y", $pastDaysTimeStamp);

        (array) $daysRange = Collection::range(0, $numberOfDays)->map(function($daysToSubtract) use ($timeStamp, $numberOfDays) {
            (integer) $timeStampOfPastDay = strtotime("-{$daysToSubtract} days",  $timeStamp);

            return new Collection([
                'day' => date('d', $timeStampOfPastDay),
                'month' => date('m', $timeStampOfPastDay),
                'year' => (integer) date('Y', $timeStampOfPastDay),
                'offset' => $numberOfDays - $daysToSubtract
            ]);
        });

        return new Collection([
            'daysAgoDate' => $daysAgoDate,
            'currentDay' => $currentDay,
            'pastDay' => $pastDay,
            'daysRange' => $daysRange,
            'pastDayMonth' => $pastDayMonth,
            'pastDayYear' => $pastDayYear,
        ]);
    }
 
    protected function getTotal(Array $set)
    {
        (object) $total = new Collection($set);

        return $total->hasKey('total')? (integer) $total->get('total') : 0;
    }

    protected function getFilteredByIdCondition()
    {
        (object) $destinationsTable = new DestinationTable;

        return "{$this->table->getName()}.destination_id IN (SELECT id FROM {$destinationsTable->getName()} destinations WHERE destinations.url_id = ?)";   
    }

}