import './AdvancedPreferencesPanel.css';

import Button from '@material-ui/core/Button';
import Card from '@material-ui/core/Card';
import CardContent from '@material-ui/core/CardContent';
import Drawer from '@material-ui/core/Drawer';
import FormControl from '@material-ui/core/FormControl';
import FormHelperText from '@material-ui/core/FormHelperText';
import InputLabel from '@material-ui/core/InputLabel';
import ListItemText from '@material-ui/core/ListItemText';
import MenuItem from '@material-ui/core/MenuItem';
import React, {Component} from 'react';
import Select from '@material-ui/core/Select';
import Typography from '@material-ui/core/Typography';

import { Callable } from '../../../../../domain/utilities/Callable';
import { HighWayPro } from '../../../../../domain/highwaypro/HighWayPro';
import {
  Observer,
} from '../../../../../domain/behaviour/events/observer/Observer';
import { Urls } from '../../../../../domain/data/finders/urls';
import Notifications from '../../../../notifications/Notifications';
import ToggleButtonGroup from '../../../../Buttons/ToggleButtonGroup';
import UpdateableField from '../../../../Fields/UpdateableField';
import UrlDestinationsLoading from './Destinations/UrlDestinationsLoading';
import _ from 'lodash';

class AdvancedPreferencesPanel extends Component {
    handleRefererChange(value) 
    {
        this.updateExtra('http_referrer_type', value)
    }

    updateExtra(name, value) 
    {
        this.props.url.loaded.extra.get(name).update(entity => {
            entity.set(name, value);
            Notifications.openLoadingNotification(true, HighWayPro.text.other.saving);
            entity.updateField(name).then(response => {
                Notifications.openLoadingNotification(false);
                Notifications.addFromResponse(response)
            }).catch(Callable.callAndReturnArgument(Notifications.addFromResponse))
        });
    }

    handleUrlUpdateOnce() 
    {
        this.props.url.loaded.extra.items.forEach(extra => {
            if (!extra.observers.length) {
                extra.add(new Observer(() => {
                    AdvancedPreferencesPanel.mountedComponent.forceUpdate();
                }));
            }
        })
    }

    render() {
        if (this.props.url.hasNotLoadedExtra()) {
        } else {
            AdvancedPreferencesPanel.mountedComponent = this;
            this.handleUrlUpdateOnce();
        }
        return (
            <Drawer classes={{paper:'hp-header-advanced-panel highwaypro', modal: 'hp-header-advanced-panel-modal'}} anchor="right" open={this.props.isOpened} onClose={this.props.onClose}>
                {this.props.url.hasNotLoadedExtra()? 
                      (<UrlDestinationsLoading title={HighWayPro.text.urls.loadingPreferences}/>)
                    : this.getCards()
                }
                <footer className="advancedFooter">
                    <Button className="deleteURL" onClick={this.tryDeletingURL.bind(this)}>
                        <span className="buttonHeader">Delete URL</span> <span className="buttonFooter">{this.props.url.name}</span>
                    </Button>
                    <p className="deleteButtonDesc">The URL <b>{this.props.url.name}</b> will be deleted permanently, including all its associated components.</p>
                </footer>
            </Drawer>
        );
    }

    getCards() 
    {
        return (
            <Card className="hp-card">
                <CardContent className="hp-card-content">
                  <div className="hp-card-header">
                    <Typography gutterBottom component="h2">
                        {HighWayPro.text.preferences.url.redirection.title}
                    </Typography>
                  </div>

                </CardContent>
                <div className="hp-card-fields --vertical">
                    <FormControl classes={{root: 'MuiFormControl-root --full-width'}}>
                        <InputLabel>
                          {HighWayPro.text.preferences.url.http_redirection_status_code.title}
                        </InputLabel>
                        <UpdateableField 
                            type="select"
                            multiple={false}
                            field="http_redirection_status_code"
                            entity={this.props.url.loaded.extra.get('http_redirection_status_code')}
                            helperText={HighWayPro.text.preferences.url.http_redirection_status_code.description}
                            renderValue={code => (
                                <div className="hp-multilabel-rendered-value">
                                    <ListItemText
                                      primary={_.findKey(
                                        HighWayPro.preferences.url._allowed.http_redirection_status_code, 
                                        _code => _code == code
                                      )}
                                      secondary={code}
                                    />
                                </div>
                            )}
                            menuItems={Object.keys(HighWayPro.preferences.url._allowed.http_redirection_status_code).map(label => {
                                    const code = HighWayPro.preferences.url._allowed.http_redirection_status_code[label];

                                    return (<MenuItem key={code} value={code}>
                                        <ListItemText
                                          primary={label}
                                          secondary={code}
                                        />
                                    </MenuItem>)
                                })}
                        />
                      </FormControl>

                    {false && (<FormControl classes={{root: 'MuiFormControl-root --full-width'}}>
                        <ToggleButtonGroup
                            value={this.props.url.loaded.extra.get('http_referrer_type').getValue('http_referrer_type')}
                            onChange={this.handleRefererChange.bind(this)}
                            label="Referrer URL to send to the target URL."
                            buttons={[
                                {
                                    value: 'unsafe-url',
                                    label: 'Always send it'
                                },
                                {
                                    value: 'origin',
                                    label: 'Only the domain'
                                },
                                {
                                    value: 'no-referrer',
                                    label: 'Hide it',
                                }
                            ]}
                        />
                        <FormHelperText classes={{root: 'MuiFormHelperText'}}>The referrer URL that is set when performing the redirection. By default, the page where the URL was clicked will be passed as the referrer URL.</FormHelperText>
                      </FormControl>)}
                </div>
            </Card>
        )
    }

    tryDeletingURL() 
    {
        if (window.confirm(HighWayPro.text.urls.delete.replace('****', this.props.url.name))) {
            this.deleteURL();
        } 
    }

    deleteURL() 
    {
        // important to store it locally here since it may change after the component has been re-rendered.
        const idToRemove = this.props.url.id;

        this.props.urls.splice(this.props.urls.findIndex(url => url.id === idToRemove), 1);

        this.props.onURLSchange({
            activeUrl: null,
            urls: this.props.urls
        }, () => {
            Notifications.openLoadingNotification(true, 'Deleting URL.');
            (new Urls).removeWithId(idToRemove)
                    .then(response => {
                        Notifications.openLoadingNotification(false);
                        Notifications.addFromResponse(response);
                    })
                    .catch(Callable.callAndReturnArgument(Notifications.addFromResponse));
        });
    }
}

export default AdvancedPreferencesPanel;